#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from typing import List

from deltafi.result import IngressResult, IngressResultItem, IngressStatusEnum
from .assertions import *
from .framework import TestCaseBase, ActionTest, IOContent


class TimedIngressTestCase(TestCaseBase):
    def __init__(self, fields: Dict):
        super().__init__(fields)
        self.memo = None
        self.results = []
        self.execute_immediate = False
        self.status = IngressStatusEnum.HEALTHY
        self.status_message = None

    def expect_ingress_result(self,
                              memo: str = None,
                              exec_immed: bool = False,
                              status: IngressStatusEnum = IngressStatusEnum.HEALTHY,
                              status_message: str = None):
        self.expected_result_type = IngressResult
        self.memo = memo
        self.execute_immediate = exec_immed
        self.status = status
        self.status_message = status_message

    def add_ingress_result_item(self, content: List[IOContent], metadata: Dict, name: str = None, annotations: Dict = None):
        if annotations is None:
            annotations = {}
        self.results.append(
            {
                'content': content,
                'metadata': metadata,
                'name': name,
                'annotations': annotations
            }
        )


class TimedIngressActionTest(ActionTest):
    def __init__(self, package_name: str):
        """
        Provides structure for testing DeltaFi TimedIngress action
        Args:
            package_name: name of the actions package for finding resources
        """
        super().__init__(package_name)

    def ingress(self, test_case: TimedIngressTestCase):
        if test_case.expected_result_type == IngressResult:
            self.expect_ingress_result(test_case)
        else:
            super().execute(test_case)

    def expect_ingress_result(self, test_case: TimedIngressTestCase):
        result = super().run_and_check_result_type(test_case, IngressResult)
        self.assert_ingress_result(test_case, result)

    def assert_ingress_result(self, test_case: TimedIngressTestCase, result: IngressResult):
        assert_equal_short(test_case.memo, result.memo, "invalid memo")
        assert_equal_short(test_case.execute_immediate, result.execute_immediate, "invalid execute_immediate")
        assert_equal_short(test_case.status, result.status, "invalid status")
        assert_equal_with_label(test_case.status_message, result.status_message, "invalid status_message")

        assert_equal_len_with_label(test_case.results, result.ingress_result_items, "item count mismatch")
        for index, ingress_item in enumerate(result.ingress_result_items):
            self.compare_one_ingress_item(test_case, ingress_item, index)
            expected = test_case.results[index]
            if 'name' in expected:
                assert_equal_with_label(expected["name"], ingress_item.delta_file_name, f"name[{index}]")

    def compare_one_ingress_item(self, test_case: TimedIngressTestCase, result: IngressResultItem, index: int):
        expected = test_case.results[index]

        # Check output
        self.compare_content_list(test_case.compare_tool, expected['content'], result.content)

        # Check metadata
        assert_keys_and_values(expected['metadata'], result.metadata)

        # Check annotations
        assert_keys_and_values(expected['annotations'], result.annotations)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import re
from abc import ABC
from abc import abstractmethod
from itertools import repeat

from deepdiff import DeepDiff, DeepSearch

from .assertions import *


class CompareHelper(ABC):
    @abstractmethod
    def compare(self, expected: str, actual: str, label: str):
        pass


class GenericCompareHelper(CompareHelper):
    def compare(self, expected: str, actual: str, label: str):
        assert_equal_with_label(expected, actual, label)


class JsonCompareHelper(CompareHelper):
    """Provides helper functions for comparing JSON/dict objects.

            Are these two JSON/dict objects equivalent?
                - compare(...)

            Are these two lists equivalent?
                - compare_lists(...)

            Is this list a subset/superset of that list?
                - compare_lists_subset(...)
                - compare_lists_superset(...)

            Select a list of values from an existing list and put them into a new list to facilitate list comparisons:
                - create_list_from_list_using_filter_regex(...)

            Is this value found (or not) in this JSON/dict object?
                - is_found(...)
                - is_not_found(...)
        """

    def __init__(self, regex_exclusion_list=None, ignore_order=True):
        """Creates and configures a JsonCompareHelper object.  If the optional 'ignore_order' is true, then the order of
                data is ignored when checking else order is enforced.  The optional 'regex_exclusion_list' excludes
                paths within the object from comparison; if empty or not provided, then no excludes are applied."""
        if regex_exclusion_list is None:
            regex_exclusion_list = []
        self.excludes = regex_exclusion_list
        self.ignore_order = ignore_order

    def __perform_find(self, obj: object, item):
        """Returns a dict of matches of the 'item' in the object 'obj'.  Both keys and values of dicts are included in
                the search.  The item may be compiled regex pattern or a string that compiles to a regex pattern.  The
                returned dict is empty if there are no matches. Excludes path(s) determined by the constructor."""
        return DeepSearch(obj, item, verbose_level=2, exclude_regex_paths=self.excludes, use_regexp=True)

    def is_not_found(self, obj: object, item):
        """Returns None if there are no occurrences of 'item' in object 'obj' else raises a ValueError.  Both keys and
                values of dicts are included in the search.  If 'item' is a list, then all elements of item must not be
                found in list, else a ValueError is raised.  The argument 'item' may be a compiled regex pattern, a
                string that compiles to a regex pattern, or a list of either or both.  Excludes path(s) and failure on
                ordering of elements are determined by the constructor."""

        all_matches = []

        if isinstance(item, list):
            for value in item:
                matches = self.__perform_find(obj, value)
                if len(matches) > 0:
                    all_matches.append(matches)
        else:
            matches = self.__perform_find(obj, item)
            if len(matches) > 0:
                all_matches.append(matches)

        if len(all_matches) > 0:
            raise ValueError("Matches found for items '" + f"{all_matches}" + "'")

        assert len(all_matches) == 0

    def is_found(self, obj: object, item):
        """Returns None if 'item' occurs in object 'obj' else raises a ValueError.  Both keys and values of dicts are
                included in the search.  If 'item' is a list, then all elements of item must occur in the object else a
                ValueError is returned.  The argument 'item' may be a compiled regex pattern, a string that compiles to
                a regex pattern, or a list of either or both. Excludes path(s) and failure on ordering of elements are
                determined by the constructor."""

        not_found_items = []

        if isinstance(item, list):
            for value in item:
                matches = self.__perform_find(obj, value)
                if len(matches) == 0:
                    not_found_items.append(value)
        else:
            matches = self.__perform_find(obj, item)
            if len(matches) == 0:
                not_found_items.append(item)

        if len(not_found_items) > 0:
            raise ValueError("No matches found for items '" + f"{not_found_items}" + "'")

        assert len(not_found_items) == 0

    def __perform_diff(self, expected, actual):
        """Returns a dict with differences between 'expected' and 'actual'.  The returned dict is empty if 'expected'
                and 'actual' are equivalent.  Both 'expected' and 'actual' must be dicts. Excludes path and failure on
                ordering of elements are determined by the constructor.  Elements must match number of repetitions."""
        return DeepDiff(expected, actual, ignore_order=self.ignore_order, report_repetition=True,
                        exclude_regex_paths=self.excludes)

    def __perform_diff_with_eval(self, expected, actual):
        """Returns None if 'expected' and 'actual' are equivalent else returns a ValueError. Both 'expected' and
                'actual' must be dicts.  Excludes path and failure on ordering of elements are determined by the
                constructor.  Elements must match number of repetitions."""

        diffs = self.__perform_diff(expected, actual)

        if len(diffs) > 0:
            raise ValueError(f"{diffs}")

        assert len(diffs) == 0

    def compare(self, expected, actual, label: str):
        """Returns None if 'expected' and 'actual' are equivalent else returns a ValueError. Both 'expected' and
                'actual' must be either dicts or strings that parse as JSON to dicts. Excludes path and failure on
                ordering of elements are determined by the constructor.  Elements must match number of repetitions."""

        if isinstance(expected, str):
            exp = json.loads(expected)
        else:
            exp = expected

        if isinstance(actual, str):
            act = json.loads(actual)
        else:
            act = actual

        return self.__perform_diff_with_eval(exp, act)

    def compare_lists(self, expected, actual: list, label: str):
        """Returns None if 'actual' is equivalent to 'expected' else returns a ValueError.

                The 'actual' argument must be a list.  The 'expected' argument may be a list or a dict.  If a list, then
                'expected' and 'actual' are compared against each other.  If a dict, then 'actual' is equivalent if it
                contains elements with the same repetitions as defined in the 'expected' dict with the key equal to the
                 element in 'actual' and the value equal to the number of repetitions.

                Order of elements is ignored.  Elements must match number of repetitions."""

        expected_list = []

        if isinstance(expected, dict):
            for key, value in expected.items():
                expected_list += list(repeat(key, value))
        else:
            expected_list = expected

        return self.__perform_diff_with_eval({"json-compare-helper-internal-list": expected_list},
                                             {"json-compare-helper-internal-list": actual})

    @staticmethod
    def compare_lists_subset(expected_subset, actual: list):
        """Returns None if the 'actual' list contains at least 'expected_subset' else returns a ValueError.  The
                'actual' list may contain 0 or more additional elements than defined in 'expected_subset'.

                The 'actual' argument must be a list.  The argument 'expected_subset' may be a list or a dict.  In the
                latter case, the key defines the item that must appear in 'actual' and the value defines the number of
                 repetitions.

                Order of elements is ignored.  Elements must match number of repetitions."""

        expected_subset_map = {}

        if isinstance(expected_subset, list):
            for item in expected_subset:
                value = expected_subset_map.get(item)
                if value is None:
                    expected_subset_map[item] = 1
                else:
                    expected_subset_map[item] = value + 1
        else:
            expected_subset_map = expected_subset

        actual_map = {}

        for item in actual:
            value = actual_map.get(item)
            if value is None:
                actual_map[item] = 1
            else:
                actual_map[item] = value + 1

        for key, value in expected_subset_map.items():
            actual_value = actual_map.get(key)
            if actual_value is None:
                raise ValueError("Actual list did not contain element '" + str(key) + "'")
            else:
                if actual_value != value:
                    raise ValueError("Actual list had item '" + str(key) + "' with repetition " + str(actual_value)
                                     + " but required repetition " + str(value))

        return None

    @staticmethod
    def compare_lists_superset(expected_superset, actual: list):
        """Returns None if the 'actual' list contains only elements that appear in 'expected_superset' else returns a
                ValueError.  The 'actual' list cannot contain more elements than are defined in 'expected_superset'; the
                'expected_superset' may define 0 or more values not contained in 'actual'.

                The 'actual' argument must be a list.  The argument 'expected_superset' may be a list or a dict.  In the
                latter case, the key defines the item that must appear in 'actual' and the value defines the number of
                 repetitions.

                Order of elements is ignored.  Elements must match number of repetitions."""

        expected_superset_map = {}

        if isinstance(expected_superset, list):
            for item in expected_superset:
                value = expected_superset_map.get(item)
                if value is None:
                    expected_superset_map[item] = 1
                else:
                    expected_superset_map[item] = value + 1
        else:
            expected_superset_map = expected_superset

        actual_map = {}

        for item in actual:
            value = actual_map.get(item)
            if value is None:
                actual_map[item] = 1
            else:
                actual_map[item] = value + 1

        for key, value in actual_map.items():
            expected_value = expected_superset_map.get(key)
            if expected_value is None:
                raise ValueError("Actual list contained element '" + str(key)
                                 + "' that did not appear in the expected superset")
            else:
                if expected_value != value:
                    raise ValueError("Actual list had item '" + str(key) + "' with repetition " + str(value)
                                     + " but required repetition " + str(expected_value))

        return None

    @staticmethod
    def create_list_from_list_using_filter_regex(regex, source_list: list):
        """Creates and returns a list of those items in the 'source_list' matching the regex in the 'regex'.  The
                argument 'regex' may be a string or a compiled regex object.

                For the 'regex', consider using anchors to explicitly match from the start (or end) of a string.  For
                example, a string such as "^malware--" will explicitly match from the start of a string."""

        def search_funct(element):

            if isinstance(regex, str):
                # 'regex' is a string
                match = re.search(regex, element)
            else:
                # assume 'regex' is a compiled regex pattern object
                match = regex.match(element)

            if match is None:
                return False
            else:
                return True

        return list(filter(search_funct, source_list))




#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import uuid
from abc import ABC
from importlib.resources import files
from typing import List

from deltafi.domain import DeltaFileMessage, Event, Content, Context
from deltafi.logger import get_logger
from deltafi.metric import Metric
from deltafi.result import ErrorResult, FilterResult
from deltafi.storage import Segment
from .assertions import *
from .compare_helpers import GenericCompareHelper, CompareHelper
from .constants import *


class IOContent:
    """
    The IOContent class holds the details for loading input or output
     content into the test framework.
    Attributes:
        file_name (str)    : The name of file in test/data.
        content_name (str) : The name of the content.
        content_type (str) : The media type of the content
        offset (int)       : Offset to use in Segment
        content_bytes (str): Optional.  If set to a String of length greater than zero, indicates to consumers of this
                             IOContent that they should bypass file read and use these bytes for content.
        no_content (bool)  : Optional.  If 'True', then consumers should not attempt to interpret content but should
                             apply other aspects of this IOContent.  When 'True', 'content_bytes' should be ignored and
                             loaded content, if any, should be interpreted as empty String or otherwise as documented by
                             the consumer.
    """

    def __init__(self, file_name: str, content_name: str = None, content_type: str = None, offset: int = 0,
                 content_bytes: str = "", no_content: bool = False):
        self.file_name = file_name
        if content_name is None:
            self.content_name = file_name
        else:
            self.content_name = content_name
        if content_type is None:
            self.content_type = IOContent.file_type(file_name)
        else:
            self.content_type = content_type
        self.offset = offset
        self.no_content = no_content
        if no_content:
            self.content_bytes = None
        else:
            self.content_bytes = content_bytes
        self.segment_uuid = uuid.uuid4()

    @classmethod
    def file_type(cls, name: str):
        if name.endswith(".json"):
            return "application/json"
        elif name.endswith(".xml"):
            return "application/xnl"
        elif name.endswith(".txt"):
            return "text/plain"
        else:
            return "application/octet-stream"


class LoadedContent:
    def __init__(self, did: str, ioc: IOContent, data: str):
        self.name = ioc.content_name
        self.content_type = ioc.content_type
        self.offset = ioc.offset
        if data is not None:
            self.data = data
        else:
            if ioc.no_content:
                self.data = ""
            else:
                self.data = ioc.content_bytes
        self.segment = Segment.from_dict(
            {"uuid": str(ioc.segment_uuid), "offset": self.offset, "size": len(self.data), "did": did})


class InternalContentService:
    def __init__(self):
        self.loaded_content = {}
        self.outputs = {}

    def load(self, content_list: List[LoadedContent]):
        for c in content_list:
            self.loaded_content[c.segment.uuid] = c

    def put_str(self, did: str, string_data: str):
        segment = Segment(uuid=str(uuid.uuid4()), offset=0, size=len(string_data), did=did)
        self.outputs[segment.uuid] = string_data
        return segment

    def get_str(self, segments: List[Segment]):
        # TODO: String multiple segment ids together
        seg_id = segments[0].uuid
        return self.loaded_content[seg_id].data

    def get_bytes(self, segments: List[Segment]):
        seg_id = segments[0].uuid
        return self.loaded_content[seg_id].data.encode('utf-8')

    def get_output(self, seg_id: str):
        if seg_id in self.outputs:
            return self.outputs[seg_id]
        elif seg_id in self.loaded_content:
            return self.loaded_content[seg_id].data
        else:
            return None


class TestCaseBase(ABC):
    def __init__(self, data: Dict):
        """
        A test case for DeltaFi python actions
        :param data: Dict of test case fields
        - action: instance of the action being tested
        - data_dir: str: subdirectory name (e.g., test name) for locating test data files, i.e., test/data/{data_dir)
        - compare_tool: (optional) CompareHelper instanced for comparing output content
        - inputs: (optional) List[IOContent]: input content to action
        - parameters: (optional) Dict: map of action input parameters
        - in_memo: (optional) str: Input 'memo' value for a TimedIngress context
        - in_meta: (optional) Dict: map of metadata as input to action
        - join_meta: (optional): List[Dict]: When a List is provided, this enables the JOIN portion of an action.
        When using JOIN, join_meta must match the size of inputs, though the Dict can be empty
        - did: (optional): str: overrides random DID
        """
        if "action" in data:
            self.action = data["action"]
        else:
            raise ValueError("action is required")

        if "data_dir" in data:
            self.data_dir = data["data_dir"]
        else:
            raise ValueError("data_dir is required")

        if "compare_tool" in data:
            self.compare_tool = data["compare_tool"]
        else:
            self.compare_tool = GenericCompareHelper()

        self.inputs = data["inputs"] if "inputs" in data else []
        self.file_name = data["file_name"] if "file_name" in data else "filename"
        self.parameters = data["parameters"] if "parameters" in data else {}
        self.in_meta = data["in_meta"] if "in_meta" in data else {}
        self.in_memo = data["in_memo"] if "in_memo" in data else None
        self.use_did = data["did"] if "did" in data else None
        self.expected_result_type = None
        self.err_or_filt_cause = None
        self.err_or_filt_context = None
        self.err_or_filt_annotations = None
        self.join_meta = data["join_meta"] if "join_meta" in data else None
        self.expected_metrics = []

    def add_metric(self, metric: Metric):
        self.expected_metrics.append(metric)

    def expect_error_result(self, cause: str, context: str, annotations: Dict = None):
        """
        A Sets the expected output of the action to an Error Result
        :param cause: the expected error cause
        :param context: the expected error context
        :param annotations: Dict: (Optional) the expected annotations
        """
        self.expected_result_type = ErrorResult
        self.err_or_filt_cause = cause
        self.err_or_filt_context = context
        self.err_or_filt_annotations = annotations

    def expect_filter_result(self, cause: str, context: str = None, annotations: Dict = None):
        """
        A Sets the expected output of the action to a Filter Result
        :param cause: the expected filter cause (message)
        :param context: (Optional) the expected filter context
        :param annotations: Dict: (Optional) the expected annotations
        """
        self.expected_result_type = FilterResult
        self.err_or_filt_cause = cause
        self.err_or_filt_context = context
        self.err_or_filt_annotations = annotations


class ActionTest(ABC):
    def __init__(self, package_name: str):
        """
        Provides structure for testing DeltaFi actions
        Args:
            package_name: name of the actions package for finding resources
        """
        self.content_service = InternalContentService()
        self.did = ""
        self.loaded_inputs = []
        self.package_name = package_name
        self.res_path = ""
        self.context = None

    def __reset__(self, did: str):
        self.content_service = InternalContentService()
        if did is None:
            self.did = str(uuid.uuid4())
        else:
            self.did = did
        self.loaded_inputs = []
        self.res_path = ""
        self.context = None

    def load_file(self, ioc: IOContent):
        file_res = self.res_path.joinpath(ioc.file_name)
        with file_res.open("r") as f:
            contents = f.read()
        return contents

    def get_contents(self, test_case: TestCaseBase):
        pkg_path = files(self.package_name)
        self.res_path = pkg_path.joinpath(f"test/data/{test_case.data_dir}/")

        # Load inputs
        for input_ioc in test_case.inputs:
            if not input_ioc.no_content and len(input_ioc.content_bytes) == 0:
                self.loaded_inputs.append(LoadedContent(self.did, input_ioc, self.load_file(input_ioc)))
            else:
                self.loaded_inputs.append(LoadedContent(self.did, input_ioc, None))

    def make_content_list(self):
        content_list = []
        for loaded_input in self.loaded_inputs:
            c = Content(name=loaded_input.name, segments=[loaded_input.segment], media_type=loaded_input.content_type,
                        content_service=self.content_service)
            content_list.append(c)
            loaded_input.content = c

        return content_list

    def make_df_msgs(self, test_case: TestCaseBase):
        content_list = self.make_content_list()
        self.content_service.load(self.loaded_inputs)

        delta_file_messages = []

        if test_case.join_meta is None:
            delta_file_messages.append(DeltaFileMessage(metadata=test_case.in_meta, content_list=content_list))
        else:
            for index, content in enumerate(content_list):
                delta_file_messages.append(DeltaFileMessage(
                    metadata=test_case.join_meta[index],
                    content_list=[content]))

        return delta_file_messages

    def make_context(self, test_case: TestCaseBase):
        action_name = INGRESS_FLOW + "." + test_case.action.__class__.__name__
        join = {} if test_case.join_meta else None
        self.context = Context(
            did=self.did,
            delta_file_name=test_case.file_name,
            data_source="DATASRC",
            flow_name=INGRESS_FLOW,
            flow_id="FLOWID",
            action_name=action_name,
            action_version="1.0",
            hostname=HOSTNAME,
            system_name=SYSTEM,
            content_service=self.content_service,
            saved_content=[],
            join=join,
            memo=test_case.in_memo,
            logger=get_logger())
        return self.context

    def make_event(self, test_case: TestCaseBase):
        return Event(delta_file_messages=self.make_df_msgs(test_case), context=self.make_context(test_case),
                     params=test_case.parameters, queue_name="", return_address="")

    def call_action(self, test_case: TestCaseBase):
        self.get_contents(test_case)
        return test_case.action.execute_action(self.make_event(test_case))

    def run_and_check_result_type(self, test_case: TestCaseBase, result_type):
        self.__reset__(test_case.use_did)
        result = self.call_action(test_case)

        if not isinstance(result, result_type):
            raise ValueError(f"Result type {result.__class__.__name__} does not match {result_type.__name__}")

        return result

    def execute_error(self, test_case: TestCaseBase):
        result = self.run_and_check_result_type(test_case, ErrorResult)
        resp = result.response()
        assert_equal_with_label(test_case.err_or_filt_cause, resp['cause'], "error cause")
        if test_case.err_or_filt_context is not None:
            assert_equal_with_label(test_case.err_or_filt_context, resp['context'], "error context")
        if test_case.err_or_filt_annotations is not None:
            assert_keys_and_values(test_case.err_or_filt_annotations, result.annotations)

    def execute_filter(self, test_case: TestCaseBase):
        result = self.run_and_check_result_type(test_case, FilterResult)
        resp = result.response()
        assert_equal_with_label(test_case.err_or_filt_cause, resp['message'], "filter cause")
        if test_case.err_or_filt_context is not None:
            assert_equal_with_label(test_case.err_or_filt_context, resp['context'], "filter context")
        if test_case.err_or_filt_annotations is not None:
            assert_keys_and_values(test_case.err_or_filt_annotations, result.annotations)

    def execute(self, test_case: TestCaseBase):
        if test_case.expected_result_type == ErrorResult:
            self.execute_error(test_case)
        elif test_case.expected_result_type == FilterResult:
            self.execute_filter(test_case)
        else:
            raise ValueError(f"unknown type: {test_case.expected_result_type}")

    @staticmethod
    def compare_content_details(expected: LoadedContent, actual: Content):
        assert_equal(expected.content_type, actual.media_type)
        assert_equal(expected.name, actual.name)

    def compare_one_content(self, comparator: CompareHelper, expected: LoadedContent, actual, index):
        self.compare_content_details(expected, actual)
        seg_id = actual.segments[0].uuid
        comparator.compare(expected.data, self.content_service.get_output(seg_id), f"Content[{index}]")

    def compare_content_list(self, comparator: CompareHelper, expected_outputs: List[IOContent], content: List):
        assert_equal_len(expected_outputs, content)
        for index, expected_ioc in enumerate(expected_outputs):
            if not expected_ioc.no_content and len(expected_ioc.content_bytes) == 0:
                expected = LoadedContent(self.did, expected_ioc, self.load_file(expected_ioc))
            else:
                expected = LoadedContent(self.did, expected_ioc, None)
            self.compare_one_content(comparator, expected, content[index], index)

    @staticmethod
    def compare_one_metric(expected: Metric, result: Metric):
        assert_equal_short(expected.name, result.name, "invalid metric name")
        assert_equal_short(expected.value, result.value, f"invalid metric value for {expected.name}")
        assert_keys_and_values(expected.tags, result.tags)

    def compare_metrics(self, expected_metrics: List[Metric], results: List[Metric]):
        if len(expected_metrics) > 0:
            assert_equal_len_with_label(expected_metrics, results, "invalid metrics count")
            for index, expected in enumerate(expected_metrics):
                self.compare_one_metric(expected, results[index])

    def has_saved_content__size(self, count: int):
        assert_equal_with_label(
            count, len(self.context.saved_content), "savedContent")



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import importlib
import inspect
import json
import os
import pkgutil
import sys
import threading
import time
import traceback
from datetime import datetime, timezone, timedelta
from importlib import metadata
from os.path import isdir, isfile, join
from pathlib import Path
from typing import List, NamedTuple

import requests
import yaml

from deltafi.action import Action, Join
from deltafi.actioneventqueue import ActionEventQueue
from deltafi.domain import Event, ActionExecution
from deltafi.exception import ExpectedContentException, MissingMetadataException
from deltafi.logger import get_logger
from deltafi.result import ErrorResult, IngressResult, TransformResult, TransformResults
from deltafi.storage import ContentService


def _coordinates():
    return PluginCoordinates(os.getenv('PROJECT_GROUP'), os.getenv('PROJECT_NAME'), os.getenv('PROJECT_VERSION'))


def _valid_file(filename: str):
    return isfile(filename) and \
        (filename.endswith(".json")
         or filename.endswith(".yaml")
         or filename.endswith(".yml"))


def _read_valid_files(path: str):
    """
    Read the contents of a directory, and returns a filtered list of files
    that can be read/parsed for plugin usage, and ignores everything else.
    :param path: name of the directory to scan
    :return: list of filtered, parsable files
    """
    files = []
    if isdir(path):
        files = [f for f in os.listdir(path) if _valid_file(join(path, f))]
    return files


def _load_resource(path: str, filename: str):
    """
    Read the content of a JSON or YAML file, and return a Python
    object of its contents, typically as a dict or list.
    To avoid exceptions, use only files returned by _read_valid_files().
    :param path: directory which contains the file to load
    :param filename: name of the file to load
    :return: dict or list of file contents
    """
    with open(join(path, filename)) as file_in:
        if filename.endswith(".json"):
            return json.load(file_in)
        elif filename.endswith(".yaml") or filename.endswith(".yml"):
            results = []
            yaml_docs = yaml.safe_load_all(file_in)
            for doc_iter in yaml_docs:
                # yaml_docs must be iterated
                results.append(doc_iter)
            if len(results) == 1:
                # Single document YAML file
                return results[0]
            else:
                # Multi-document YAML file
                return results
    raise RuntimeError(f"File type not supported: {filename}")


def _load__all_resource(path: str, file_list: List[str]):
    resources = []
    for f in file_list:
        r = _load_resource(path, f)
        if isinstance(r, list):
            resources.extend(r)
        else:
            resources.append(r)
    return resources


def _find_variables_filename(names: List[str]):
    if 'variables.json' in names:
        return 'variables.json'
    elif 'variables.yaml' in names:
        return 'variables.yaml'
    elif 'variables.yml' in names:
        return 'variables.yml'
    else:
        return None


def _setup_queue(max_connections):
    url = os.getenv('VALKEY_URL', 'http://deltafi-valkey-master:6379')
    password = os.getenv('VALKEY_PASSWORD')
    app_name = os.getenv('APP_NAME')
    return ActionEventQueue(url, max_connections, password, app_name)


def _setup_content_service():
    minio_url = os.getenv('MINIO_URL', 'http://deltafi-minio:9000')
    return ContentService(minio_url,
                          os.getenv('MINIO_ACCESSKEY'),
                          os.getenv('MINIO_SECRETKEY'))


class ActionThread(object):
    def __init__(self, clazz: Action, thread_num: int, name: str, execution: ActionExecution = None):
        self.clazz = clazz
        self.thread_num = thread_num
        self.name = name
        self.execution = execution

    def logger_name(self):
        return f"{self.name}#{self.thread_num}"


class PluginCoordinates(object):
    def __init__(self, group_id: str, artifact_id: str, version: str):
        self.group_id = group_id
        self.artifact_id = artifact_id
        self.version = version

    def __json__(self):
        return {
            "groupId": self.group_id,
            "artifactId": self.artifact_id,
            "version": self.version
        }


LONG_RUNNING_TASK_DURATION = timedelta(seconds=5)


class Plugin(object):
    def __init__(self, description: str, plugin_name: str = None, plugin_coordinates: PluginCoordinates = None,
                 actions: List = None, action_package: str = None,
                 thread_config: dict = None):
        """
        Initialize the plugin object
        :param plugin_name: Name of the plugin project
        :param description: Description of the plugin
        :param plugin_coordinates: plugin coordinates of the plugin, if None the coordinates must be defined in
                                   environment variables
        :param actions: list of action classes to run
        :param action_package: name of the package containing the actions to run
        :param  thread_config: map of action class name and thread count. Actions not found default to 1 thread.
        """
        self.logger = get_logger()

        self.content_service = None
        self.queue = None
        self.singleton_actions = []
        self.action_threads = []
        self.thread_config = {}
        if thread_config is not None:
            self.thread_config = thread_config
        self.core_url = os.getenv('CORE_URL')
        self.image = os.getenv('IMAGE')
        self.image_pull_secret = os.getenv('IMAGE_PULL_SECRET')
        action_classes = []
        if actions is not None and len(actions):
            action_classes.extend(actions)

        if action_package is not None:
            found_actions = Plugin.find_actions(action_package)
            if len(found_actions):
                action_classes.extend(found_actions)

        unique_actions = dict.fromkeys(action_classes)
        self.singleton_actions = [action() for action in unique_actions]

        self.description = description
        self.display_name = os.getenv('PROJECT_NAME') if plugin_name is None else plugin_name
        self.coordinates = _coordinates() if plugin_coordinates is None else plugin_coordinates

        if os.getenv('ACTIONS_HOSTNAME'):
            self.hostname = os.getenv('ACTIONS_HOSTNAME')
        elif os.getenv('HOSTNAME'):
            self.hostname = os.getenv('HOSTNAME')
        elif os.getenv('COMPUTERNAME'):
            self.hostname = os.getenv('COMPUTERNAME')
        else:
            self.hostname = 'UNKNOWN'

        self.logger.debug(f"Initialized ActionRunner with actions {self.singleton_actions}")

    @staticmethod
    def find_actions(package_name) -> List[object]:
        """
        Find all concrete classes that extend the base Action class in the given package
        :param package_name: name of the package to load and scan for actions
        :return: list of classes that extend the Action class
        """
        package = importlib.import_module(package_name)
        classes = []
        visited = set()

        # Iterate over all submodules in the package
        for _, module_name, _ in pkgutil.walk_packages(package.__path__):
            try:
                module = importlib.import_module(package.__name__ + '.' + module_name)
            except ModuleNotFoundError:
                continue

            # Iterate over all members in the module
            for name, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and obj.__module__.startswith(package_name) and obj not in visited:
                    if Plugin.is_action(obj):
                        classes.append(obj)
                    visited.add(obj)

        return classes

    @staticmethod
    def is_action(maybe_action: type) -> bool:
        """
        Check if the given object is a non-abstract subclass of the Action class
        :param maybe_action: object to inspect to see if it is an Action class
        :return: true if the object is a non-abstract subclass of the Action class
        """
        return not inspect.isabstract(maybe_action) and issubclass(maybe_action, Action)

    def action_name(self, action):
        return f"{self.coordinates.group_id}.{action.__class__.__name__}"

    def _action_json(self, action):
        return {
            'name': self.action_name(action),
            'type': action.action_type.name,
            'supportsJoin': isinstance(action, Join),
            'schema': action.param_class().model_json_schema(),
            'actionOptions': action.action_options.json()
        }

    @staticmethod
    def load_integration_tests(tests_path: str):
        test_files = _read_valid_files(tests_path)
        return _load__all_resource(tests_path, test_files)

    @staticmethod
    def load_variables(flows_path: str, flow_files: List[str]):
        variables = []
        variables_filename = _find_variables_filename(flow_files)
        if variables_filename is not None:
            flow_files.remove(variables_filename)
            variables = _load__all_resource(flows_path, [variables_filename])
        return variables

    def registration_json(self):
        flows_path = str(Path(os.path.dirname(os.path.abspath(sys.argv[0]))) / 'flows')
        tests_path = str(Path(os.path.dirname(os.path.abspath(sys.argv[0]))) / 'integration')

        variables = []
        flow_files = _read_valid_files(flows_path)
        if len(flow_files) == 0:
            self.logger.warning(
                f"Flows directory ({flows_path}) does not exist or contains no valid files. No flows will be installed.")
        else:
            variables = self.load_variables(flows_path, flow_files)

        flows = _load__all_resource(flows_path, flow_files)
        actions = [self._action_json(action) for action in self.singleton_actions]

        test_files = self.load_integration_tests(tests_path)
        if len(test_files) == 0:
            self.logger.warning(
                f"tests directory ({tests_path}) does not exist or contains no valid files. No tests will be installed.")

        return {
            'pluginCoordinates': self.coordinates.__json__(),
            'displayName': self.display_name,
            'description': self.description,
            'actionKitVersion': metadata.version('deltafi'),
            'image': self.image,
            'imagePullSecret': self.image_pull_secret,
            'dependencies': [],
            'actions': actions,
            'variables': variables,
            'flowPlans': flows,
            'integrationTests': test_files
        }

    def _register(self):
        url = f"{self.core_url}/plugins"
        headers = {'Content-type': 'application/json'}
        registration_json = self.registration_json()

        self.logger.info(f"Registering plugin:\n{registration_json}")

        response = requests.post(url, headers=headers, json=registration_json)
        if not response.ok:
            self.logger.error(f"Failed to register plugin ({response.status_code}):\n{response.content}")
            exit(1)

        self.logger.info("Plugin registered")

    def run(self):
        self.logger.info("Plugin starting")

        for action in self.singleton_actions:
            num_threads = 1
            if self.action_name(action) in self.thread_config:
                maybe_num_threads = self.thread_config[self.action_name(action)]
                if type(maybe_num_threads) is int and maybe_num_threads > 0:
                    num_threads = maybe_num_threads
                else:
                    self.logger.error(f"Ignoring non-int or invalid thread value {maybe_num_threads}")
            for i in range(num_threads):
                action_thread = ActionThread(action, i, self.action_name(action))
                self.action_threads.append(action_thread)

        self.queue = _setup_queue(len(self.action_threads) + 1)
        self.content_service = _setup_content_service()
        self._register()

        for action_thread in self.action_threads:
            threading.Thread(target=self._do_action, args=(action_thread,)).start()

        hb_thread = threading.Thread(target=self._heartbeat)
        hb_thread.start()

        self.logger.info("All threads running")

        f = open("/tmp/running", "w")
        f.close()

        self.logger.info("Application initialization complete")
        hb_thread.join()

    def _heartbeat(self):
        long_running_actions = set()
        while True:
            try:
                # Set heartbeats
                for action_thread in self.action_threads:
                    self.queue.heartbeat(action_thread.name)

                # Record long running tasks
                new_long_running_actions = set()
                for action_thread in self.action_threads:
                    action_execution = action_thread.execution
                    if action_execution and action_execution.exceeds_duration(LONG_RUNNING_TASK_DURATION):
                        new_long_running_actions.add(action_execution)
                        self.queue.record_long_running_task(action_execution)

                # Remove old long running tasks
                tasks_to_remove = long_running_actions - new_long_running_actions
                for action_execution in tasks_to_remove:
                    self.queue.remove_long_running_task(action_execution)

                long_running_actions = new_long_running_actions

            except Exception as e:
                self.logger.error(f"Failed to register action queue heartbeat or record long running tasks: {e}", e)
            finally:
                time.sleep(10)

    @staticmethod
    def to_response(event, start_time, stop_time, result):
        response = {
            'did': event.context.did,
            'flowName': event.context.flow_name,
            'flowId': event.context.flow_id,
            'actionName': event.context.action_name,
            'start': start_time,
            'stop': stop_time,
            'type': result.result_type,
            'metrics': [metric.json() for metric in result.metrics]
        }
        if result.result_key is not None:
            response[result.result_key] = result.response()
        return response

    def _do_action(self, action_thread: ActionThread):
        action_logger = get_logger(action_thread.logger_name())
        action_logger.info(f"Listening on {action_thread.name}")

        while True:
            try:
                event_string = self.queue.take(action_thread.name)
                event = Event.create(json.loads(event_string), self.content_service, action_logger)
                start_time = time.time()
                action_logger.debug(f"Processing event for did {event.context.did}")

                action_thread.execution = ActionExecution(action_thread.name, event.context.action_name,
                                                          action_thread.thread_num, event.context.did,
                                                          datetime.now(timezone.utc))

                try:
                    result = action_thread.clazz.execute_action(event)
                except ExpectedContentException as e:
                    result = ErrorResult(event.context,
                                         f"Action attempted to look up element {e.index + 1} (index {e.index}) from "
                                         f"content list of size {e.size}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingMetadataException as e:
                    result = ErrorResult(event.context,
                                         f"Missing metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except BaseException as e:
                    result = ErrorResult(event.context,
                                         f"Action execution {type(e)} exception", f"{str(e)}\n{traceback.format_exc()}")

                action_thread.execution = None

                response = Plugin.to_response(
                    event, start_time, time.time(), result)

                Plugin.orphaned_content_check(action_logger, event.context, result, response)

                topic = 'dgs'
                if event.return_address:
                    topic += f"-{event.return_address}"
                self.queue.put(topic, json.dumps(response))
            except BaseException as e:
                action_logger.error(f"Unexpected {type(e)} error: {str(e)}\n{traceback.format_exc()}")
                time.sleep(1)

    @staticmethod
    def orphaned_content_check(logger, context, result, response):
        if len(context.saved_content) > 0:
            to_delete = Plugin.find_unused_content(context.saved_content, result)
            if len(to_delete) > 0:
                errors = context.content_service.delete_all(to_delete)
                for e in errors:
                    logger.error(f"Unable to delete object(s), {e}")
                logger.warning(
                    f"Deleted {len(to_delete)} unused content entries for did {context.did} due to a {response['type']} event by {response['actionName']}")

    @staticmethod
    def find_unused_content(saved_content, result):
        segments_in_use = Plugin.used_segment_names(result)
        saved_segments = Plugin.get_segment_names(saved_content)
        to_delete = []
        for key, value in saved_segments.items():
            if key not in segments_in_use:
                to_delete.append(value)
        return to_delete

    @staticmethod
    def used_segment_names(result):
        segment_names = {}
        if isinstance(result, TransformResult):
            segment_names.update(result.get_segment_names())
        elif isinstance(result, TransformResults):
            segment_names.update(result.get_segment_names())
        elif isinstance(result, IngressResult):
            segment_names.update(result.get_segment_names())
        return segment_names

    @staticmethod
    def get_segment_names(content_list):
        segment_names = {}
        for content in content_list:
            segment_names.update(content.get_segment_names())
        return segment_names



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2025 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import abc
import uuid
from enum import Enum
from typing import NamedTuple

from deltafi.domain import Content, Context
from deltafi.metric import Metric


class Result:
    __metaclass__ = abc.ABCMeta

    def __init__(self, result_key, result_type, context):
        self.result_key = result_key
        self.result_type = result_type
        self.metrics = []
        self.context = context

    @abc.abstractmethod
    def response(self):
        pass

    def add_metric(self, metric: Metric):
        self.metrics.append(metric)
        return self


class EgressResult(Result):
    def __init__(self, context: Context):
        super().__init__(None, 'EGRESS', context)

    def response(self):
        return None


class ErrorResult(Result):
    def __init__(self, context: Context, error_cause: str, error_context: str):
        super().__init__('error', 'ERROR', context)
        self.error_cause = error_cause
        self.error_context = error_context
        self.annotations = {}

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'cause': self.error_cause,
            'context': self.error_context,
            'annotations': self.annotations
        }


class FilterResult(Result):
    def __init__(self, context: Context, filtered_cause: str, filtered_context: str = None):
        super().__init__('filter', 'FILTER', context)
        self.filtered_cause = filtered_cause
        self.filtered_context = filtered_context
        self.annotations = {}

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def response(self):
        return {
            'message': self.filtered_cause,
            'context': self.filtered_context,
            'annotations': self.annotations
        }


class IngressResultItem:
    def __init__(self, context: Context, delta_file_name: str):
        self.context = context
        self._did = str(uuid.uuid4())
        self.content = []
        self.metadata = {}
        self.annotations = {}
        self.delta_file_name = delta_file_name

    @property
    def did(self):
        return self._did

    # content can be a single Content or a List[Content]
    def add_content(self, content):
        if content:
            if type(content) == list:
                self.content.extend(content)
            else:
                self.content.append(content)

        return self

    def save_string_content(self, string_data: str, name: str, media_type: str, tags: set = None):
        segment = self.context.content_service.put_str(self._did, string_data)
        c = Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service)
        if tags is not None:
            c.add_tags(tags)
        self.content.append(c)
        self.context.saved_content.append(c)
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str, tags: set = None):
        segment = self.context.content_service.put_bytes(self._did, byte_data)
        c = Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service)
        if tags is not None:
            c.add_tags(tags)
        self.content.append(c)
        self.context.saved_content.append(c)
        return self

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def get_segment_names(self):
        segment_names = {}
        for c in self.content:
            segment_names.update(c.get_segment_names())
        return segment_names

    def annotate(self, key: str, value: str):
                    self.annotations[key] = value
                    return self

    def response(self):
        return {
            'did': self._did,
            'deltaFileName': self.delta_file_name,
            'metadata': self.metadata,
            'content': [content.json() for content in self.content],
            'annotations': self.annotations
        }


class IngressStatusEnum(Enum):
    HEALTHY = 'HEALTHY'
    DEGRADED = 'DEGRADED'
    UNHEALTHY = 'UNHEALTHY'


class IngressResult(Result):
    def __init__(self, context: Context):
        super().__init__('ingress', 'INGRESS', context)
        self.memo = None
        self.ingress_result_items = []
        self.execute_immediate = False
        self.status = IngressStatusEnum.HEALTHY
        self.status_message = None

    def add_item(self, ingress_result_item: IngressResultItem):
        self.ingress_result_items.append(ingress_result_item)
        return self

    def get_segment_names(self):
        segment_names = {}
        for ingress_item in self.ingress_result_items:
            segment_names.update(ingress_item.get_segment_names())
        return segment_names

    def response(self):
        return {
            'memo': self.memo,
            'executeImmediate': self.execute_immediate,
            'ingressItems': [ingress_result_item.response() for ingress_result_item in self.ingress_result_items],
            'status': self.status.value,
            'statusMessage': self.status_message
        }


class TransformResult(Result):
    def __init__(self, context: Context):
        super().__init__('transform', 'TRANSFORM', context)
        self.content = []
        self.annotations = {}
        self.metadata = {}
        self.delete_metadata_keys = []

    # content can be a single Content or a List[Content]
    def add_content(self, content):
        if content:
            if type(content) == list:
                self.content.extend(content)
            else:
                self.content.append(content)

        return self

    def save_string_content(self, string_data: str, name: str, media_type: str, tags: set = None):
        segment = self.context.content_service.put_str(self.context.did, string_data)
        c = Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service)
        if tags is not None:
            c.add_tags(tags)
        self.content.append(c)
        self.context.saved_content.append(c)
        return self

    def save_byte_content(self, byte_data: bytes, name: str, media_type: str, tags: set = None):
        segment = self.context.content_service.put_bytes(self.context.did, byte_data)
        c = Content(name=name, segments=[segment], media_type=media_type, content_service=self.context.content_service)
        if tags is not None:
            c.add_tags(tags)
        self.content.append(c)
        self.context.saved_content.append(c)
        return self

    def set_metadata(self, metadata: dict):
        self.metadata = metadata
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def annotate(self, key: str, value: str):
        self.annotations[key] = value
        return self

    def delete_metadata_key(self, key: str):
        self.delete_metadata_keys.append(key)
        return self

    def get_segment_names(self):
        segment_names = {}
        for c in self.content:
            segment_names.update(c.get_segment_names())
        return segment_names

    def json(self):
        return {
            'did': self.context.did,
            'content': [content.json() for content in self.content],
            'annotations': self.annotations,
            'metadata': self.metadata,
            'deleteMetadataKeys': self.delete_metadata_keys
        }

    def response(self):
        return [self.json()]


class ChildTransformResult(TransformResult):
    delta_file_name: str

    def __init__(self, context: Context, delta_file_name: str = None):
        super().__init__(context.child_context())
        self.delta_file_name = delta_file_name

    def json(self):
        j = super().json()
        if self.delta_file_name is not None:
            j['name'] = self.delta_file_name
        return j


class TransformResults(Result):
    def __init__(self, context: Context):
        super().__init__('transform', 'TRANSFORM', context)
        self.child_results = []

    def add_result(self, result: ChildTransformResult):
        self.child_results.append(result)
        return self

    def get_segment_names(self):
        segment_names = {}
        for child_result in self.child_results:
            segment_names.update(child_result.get_segment_names())
        return segment_names

    def response(self):
        transform_events = []
        for child_result in self.child_results:
            json_dict = child_result.json()
            transform_events.append(json_dict)
        return transform_events
